@@ -19,6 +19,7 @@ from iiif_downloader import Manifest
 from rasterfairy import coonswarp
 from keras.models import Model
 from scipy.stats import kde
+from PIL import ImageFile
 import keras.backend as K
 import tensorflow as tf
 import multiprocessing
@@ -60,8 +61,9 @@ try:
   from tf_pose.networks import get_graph_path, model_wh
   from tf_pose.estimator import TfPoseEstimator
   from tf_pose import common
+  pose_ready = True
 except:
-  pass
+  pose_ready = False
 
 try:
   from hdbscan import HDBSCAN
@@ -76,6 +78,9 @@ tf_config.gpu_options.allow_growth = True
 tf_config.log_device_placement = True
 sess = tf.compat.v1.Session(config=tf_config)
 
+# handle truncated images in PIL (managed by Pillow)
+ImageFile.LOAD_TRUNCATED_IMAGES = True
+
 '''
 NB: Keras Image class objects return image.size as w,h
     Numpy array representations of images return image.shape as h,w,c
@@ -106,6 +111,7 @@ config = {
   'plot_id': str(uuid.uuid1()),
   'seed': 24,
   'n_clusters': 12,
+  'geojson': None,
 }
 
 
@@ -187,6 +193,9 @@ def filter_images(**kwargs):
     except ValueError:
       print(' * skipping {} because it contains 0 height or width when resized'.format(i.path))
       continue
+    except OSError:
+      print(' * skipping {} because it could not be resized'.format(i.path))
+      continue
     # remove images that are too wide for the atlas
     if (w/h) > (kwargs['atlas_size']/kwargs['cell_size']):
       print(' * skipping {} because its dimensions are oblong'.format(i.path))
@@ -235,9 +244,13 @@ def get_image_paths(**kwargs):
   if os.path.exists(kwargs['images']):
     with open(kwargs['images']) as f:
       f = [i.strip() for i in f.read().split('\n') if i.strip()]
-      if [i.startswith('http') for i in f]:
-        for i in f: Manifest(url=i).save_images(limit=1)
-        image_paths = sorted(glob2.glob(os.path.join('iiif-downloads', 'images', '*')))
+      for i in f:
+        if i.startswith('http'):
+          try:
+            Manifest(url=i).save_images(limit=1)
+          except:
+            print(' * could not download url ' + i)
+      image_paths = sorted(glob2.glob(os.path.join('iiif-downloads', 'images', '*')))
   # handle case where images flag points to a glob of images
   if not image_paths:
     image_paths = sorted(glob2.glob(kwargs['images']))
@@ -388,6 +401,8 @@ def get_manifest(**kwargs):
   point_sizes['max'] = point_sizes['grid'] * 1.2
   point_sizes['scatter'] = point_sizes['grid'] * .2
   point_sizes['initial'] = point_sizes['scatter']
+  point_sizes['categorical'] = point_sizes['grid'] * 0.6
+  point_sizes['geographic'] = point_sizes['grid'] * 0.025
   # fetch the date distribution data for point sizing
   if 'date' in layouts and layouts['date']:
     date_layout = read_json(layouts['date']['labels'], **kwargs)
@@ -511,6 +526,7 @@ def get_layouts(**kwargs):
   alphabetic = get_alphabetic_layout(**kwargs)
   categorical = get_categorical_layout(**kwargs)
   date = get_date_layout(**kwargs)
+  geographic = get_geographic_layout(**kwargs)
   layouts = {
     'umap': {
       'layout': umap,
@@ -525,6 +541,7 @@ def get_layouts(**kwargs):
     'pose': pose,
     'categorical': categorical,
     'date': date,
+    'geographic': geographic,
   }
   return layouts
 
@@ -734,6 +751,9 @@ def crop_openpose_figure(im, vec, margin=0.4):
 
 def subdivide_images_with_openpose(**kwargs):
   '''Cut each input image into single-pose subimages and save vectors for each'''
+  if not pose_ready:
+    print(' * to use --extract_poses, please install tf-pose==0.11.0 and build your plot again')
+    sys.exit()
   print(' * subdividing images with OpenPose model')
   download_cmu_model()
   # determine the subset of input images for which we've already parsed pose vectors
@@ -829,6 +849,7 @@ def get_cmu_graph_path():
   '''Return the path to the location where the CMU graph will be stored'''
   return os.path.join(dirname(realpath(__file__)), 'models', 'cmu', 'graph_opt.pb')
 
+
 ##
 # Date Layout
 ##
@@ -958,10 +979,9 @@ def get_categorical_layout(null_category='Other', margin=2, **kwargs):
   # determine the out path and return from cache if possible
   out_path = get_path('layouts', 'categorical', **kwargs)
   labels_out_path = get_path('layouts', 'categorical-labels', **kwargs)
-  if os.path.exists(out_path): return out_path
   # accumulate d[category] = [indices of points with category]
   categories = [i.get('category', None) for i in kwargs['metadata']]
-  if not any(categories) or len(set(categories)) == 1: return False
+  if not any(categories) or len(set(categories) - set([None])) == 1: return False
   d = defaultdict(list)
   for idx, i in enumerate(categories): d[i].append(idx)
   # store the number of observations in each group
@@ -1023,7 +1043,7 @@ def get_categorical_boxes(group_counts, margin=2):
   y = margin
   y_spots = []
   for idx, i in enumerate(boxes):
-    if (y + i.h) < wrap:
+    if (y + i.h) <= wrap:
       y_spots.append(y)
       y += i.h+margin
   y_spot_index = 0
@@ -1069,6 +1089,46 @@ class Box:
     self.y = None if len(args) < 5 else args[4]
 
 
+##
+# Geographic Layout
+##
+
+
+def get_geographic_layout(**kwargs):
+  '''Return a 2D array of image positions corresponding to lat, lng coordinates'''
+  out_path = get_path('layouts', 'geographic', **kwargs)
+  l = []
+  coords = False
+  for idx, i in enumerate(stream_images(**kwargs)):
+    lat = float(i.metadata.get('lat', 0)) / 180
+    lng = float(i.metadata.get('lng', 0)) / 180 # the plot draws longitude twice as tall as latitude
+    if lat or lng: coords = True
+    l.append([lng, lat])
+  if coords:
+    if kwargs['geojson']:
+      process_geojson(kwargs['geojson'])
+    return {
+      'layout': write_layout(out_path, l, scale=False, **kwargs)
+    }
+  elif kwargs['geojson']:
+    print(' * GeoJSON is only processed if you also provide lat/lng coordinates for your images in a metadata file!')
+  return None
+
+
+def process_geojson(geojson_path):
+  '''Given a GeoJSON filepath, write a minimal JSON output in lat lng coordinates'''
+  with open(geojson_path, 'r') as f:
+    geojson = json.load(f)
+  l = []
+  for i in geojson:
+    if isinstance(i, dict):
+      for j in i.get('coordinates', []):
+        for k in j:
+          l.append(k)
+  with open(os.path.join('output', 'assets', 'json', 'geographic-features.json'), 'w') as out:
+    json.dump(l, out)
+
+
 ##
 # Helpers
 ##
@@ -1086,8 +1146,10 @@ def get_path(*args, **kwargs):
 
 def write_layout(path, obj, **kwargs):
   '''Write layout json `obj` to disk and return the path to the saved file'''
-  obj = (minmax_scale(obj)-0.5)*2 # scale -1:1
-  obj = round_floats(obj)
+  if kwargs.get('scale', True) != False:
+    obj = (minmax_scale(obj)-0.5)*2 # scale -1:1
+  if kwargs.get('round', True) != False:
+    obj = round_floats(obj)
   return write_json(path, obj, **kwargs)
 
 
@@ -1234,7 +1296,7 @@ class Image:
   def __init__(self, *args, **kwargs):
     self.path = args[0]
     self.original = load_img(self.path)
-    self.metadata = kwargs['metadata']
+    self.metadata = kwargs['metadata'] if kwargs['metadata'] else {}
 
   def resize_to_max(self, n):
     '''
@@ -1302,6 +1364,7 @@ def parse():
   parser.add_argument('--plot_id', type=str, default=config['plot_id'], help='unique id for a plot; useful for resuming processing on a started plot')
   parser.add_argument('--seed', type=int, default=config['seed'], help='seed for random processes')
   parser.add_argument('--n_clusters', type=int, default=config['n_clusters'], help='number of clusters to use when clustering with kmeans')
+  parser.add_argument('--geojson', type=str, default=config['geojson'], help='path to a GeoJSON file with shapes to be rendered on a map')
   config.update(vars(parser.parse_args()))
   process_images(**config)
 
